# -*- coding: utf-8 -*-
"""
Created on Fri Nov  8 16:38:42 2019

@author: zadmin
"""

# imports  of all the packages of pytorch and numpy
import torch
import matplotlib.pyplot as plt
import numpy as np
import torch.nn.functional as F
from torch import nn
from torchvision import datasets, transforms, models
import sys
import os
import math

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from semih_gradcam_script import runner
import zetane
import zetane.context as ztn

#cuda check
device = torch.device('cuda:0' if torch.cuda.is_available() else "cpu")

# convertion of images to fit the shape of the vgg model
def im_convert(tensor):
    image = tensor.cpu().clone().detach().numpy()
    image = image.transpose(1, 2, 0)
    image = image * np.array((0.5,0.5,0.5)) + np.array((0.5,0.5,0.5))
    image = image.clip(0,1)
    return image

# laoding of the training set images and test set images, resizing then, some preprocessing of the data transforms
transform_train = transforms.Compose([transforms.Resize((224, 224)),
                                     transforms.RandomHorizontalFlip(),
                                     transforms.RandomAffine(0, shear=10, scale=(0.8, 1.2)),
                                     transforms.ColorJitter(brightness=1, contrast=1, saturation=1),
                                     transforms.ToTensor(),
                                     transforms.Normalize((0.5,), (0.5,))
                                    ])
transform = transforms.Compose([transforms.Resize((224,224)),
                               transforms.ToTensor(),
                               transforms.Normalize((0.5,),(0.5,))
                               ])
training_dataset = datasets.ImageFolder('C:/Users/semik/zetane/z10/python/package/semih_gradcam/ObstacleDataset/Train', transform=transform_train)
validation_dataset = datasets.ImageFolder('C:/Users/semik/zetane/z10/python/package/semih_gradcam/ObstacleDataset/Test', transform=transform)

training_loader = torch.utils.data.DataLoader(training_dataset, batch_size=20, shuffle=True)
validation_loader = torch.utils.data.DataLoader(validation_dataset, batch_size=20, shuffle=False)

print(len(training_dataset))
print(len(validation_dataset))

# launch viz
zcontext = ztn.Context()
zcontext.launch()
zimg = zcontext.image()
zonnx = None

# classes in which the data will be classified as obstacles and non-obstacles
classes = ['Non_Obstacles','Obstacles']

# iteration of the data in the dataset to verify the images are labelled correctly in the correct folders
dataiter = iter(training_loader)
images, labels = dataiter.next()
fig = plt.figure(figsize=(25, 4))

for idx in np.arange(20):
    ax = fig.add_subplot(2, 10, idx+1, xticks=[], yticks=[])
    plt.imshow(im_convert(images[idx]))
    ax.set_title(classes[labels[idx].item()])

# Loading the pretrained vgg model from the torch vision library of pytorch
model = models.vgg16(pretrained=True)

print(model)

# using the transfer learning to use pretrained vgg to train the model on the different dataset to evaluate the required results
for param in model.features.parameters():
    param.requires_grad = False

# changing the final fully connected layer so that the dataset is evaluated on the two classes instead of the 1000 classes
import torch.nn as nn
n_inputs = model.classifier[6].in_features
last_layer = nn.Linear(n_inputs, len(classes))
#change of the layer happens at this point
model.classifier[6] = last_layer
model.to(device)
print(model)

print(model.classifier[6].out_features)

#Using the crossEntropLoss() function of the nn module of the pytorch to calculate the accuracy and the log loss
criterion = nn.CrossEntropyLoss()
# Hyper tunning the parameters to get the results
optimizer = torch.optim.Adam(model.parameters(), lr = 0.00001)

# iterations of the dataset on the model to predict the accuracy and log loss
epochs = 1
#To keep the track of the running loss and running corrects of the training set and test set
running_loss_history = []
running_corrects_history = []
val_running_loss_history = []
val_running_corrects_history = []

for e in range(epochs):
    #Initializing the running_loss and running_corrects of the training set and test set
    running_loss = 0.0
    running_corrects = 0.0
    val_running_loss = 0.0
    val_running_corrects = 0.0

    for inputs, labels in training_loader:
        inputs = inputs.to(device)
        labels = labels.to(device)
        #input image is passed to the model and result is stored in the output variable
        # To calculate loss the labels of the outputs are compared with the ground truth label. For example if i pass the image of the car and the model predict car then the prediction is correct and hence loss is less
        outputs = model(inputs)
        loss = criterion(outputs, labels)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        _, preds = torch.max(outputs, 1)
        running_loss += loss.item()
        # running_corrects is the summation of all the correct predictions made by the model. For example if out of the 30 images the correctly predicted images are 25 then running_corrects will be 25
        running_corrects += torch.sum(preds == labels.data)

        # RUN GRADCAM
        filepaths_gradcam = runner(inputs[0], model, depth=3, backprop_type="gradcam", classes=classes)
        zimg.update(filepaths_gradcam)
        #print(filepaths_gradcam)

        try:
            torch.onnx.export(model, inputs, R"Obstacle.onnx")
            # Update Zetane-side ONNX variable
            if zonnx:
                zonnx.update(filepath=R"Obstacle.onnx", inputs=R"Obstacle.npy")
            else:
                zonnx = zcontext.onnx(filepath=R"Obstacle.onnx", inputs=R"Obstacle.npy")
                zonnx.position(x=3)

        except OSError as e:
            print(e)



    else:
        # Same for the validation/Test set
        with torch.no_grad():
            for val_inputs, val_labels in validation_loader:
                val_inputs = val_inputs.to(device)
                val_labels = val_labels.to(device)
                val_outputs = model(val_inputs)
                val_loss = criterion(val_outputs, val_labels)

                _, val_preds = torch.max(val_outputs, 1)
                val_running_loss += val_loss.item()
                val_running_corrects += torch.sum(val_preds == val_labels.data)

            # To keep the loss and accuracy normalized training_loader.dataset is being used if running_loss is divided by just training_loader then the result will not be normalized hence difficult of evaluate
            # Calucaltion of loss at each iteration
            epoch_loss = running_loss/len(training_loader.dataset)
            #Calculation of accuracy at each iteration
            epoch_acc = running_corrects.float()/len(training_loader.dataset)
            running_loss_history.append(epoch_loss)
            running_corrects_history.append(epoch_acc)

            val_epoch_loss = val_running_loss/len(validation_loader.dataset)
            val_epoch_acc  = val_running_corrects.float()/len(validation_loader.dataset)
            val_running_loss_history.append(val_epoch_loss)
            val_running_corrects_history.append(val_epoch_acc)
            print('epoch:', (e+1))
            print('training loss: {:.4f}, acc {:.4f}'.format(epoch_loss, epoch_acc.item()))
            print('validation loss: {:.4f}, validation acc {:.4f}'.format(val_epoch_loss, val_epoch_acc.item()))
            #The trained model can be stored as onnx after every iteration and can overwrite the previous saved model same way it was done for Fashion Mnist
            #torch.onnx.export(model, val_inputs, 'Obstacle.onnx')
            torch.save(model.state_dict(), 'mytraining.pt')


#Saved the fully trained model instead of saving it after every iteration. It appears faster
torch.onnx.export(model, val_inputs, 'Obstacle.onnx')

# ploting the variation in the training loss and validation loss to visualize the over fitting scenario. Overfitting can also be interpreted by the huge difference in the accuracies of the training set and test set.
plt.plot(running_loss_history, label = 'training loss')
plt.plot(val_running_loss_history, label = 'validation loss')
plt.legend()

#ploting the variation in the training accuracy and test accuracy
#For example if the difference in the accuracy is around 15-20% from training set and test set then there is a huge possibility of overfitting.
#Depends on the quality of the data
plt.plot(running_corrects_history, label = 'training accuracy')
plt.plot(val_running_corrects_history, label = 'validation accuracy')
plt.legend()

import PIL.ImageOps

#Testing the trained model on the test image
import requests
from PIL import Image

url = '000002.jpg'
img = Image.open(url)
plt.imshow(img)

img = transform(img)
plt.imshow(im_convert(img))

image = img.to(device).unsqueeze(0)
output = model(image)
_, pred = torch.max(output, 1)
print(classes[pred.item()])

#Testing the model on the batch of the images from the test set
dataiter = iter(validation_loader)
images, labels = dataiter.next()
images = images.to(device)
labels = labels.to(device)
output = model(images)
_, preds = torch.max(output, 1)

fig = plt.figure(figsize=(25, 4))

for idx in np.arange(20):
    ax = fig.add_subplot(2, 10, idx+1, xticks=[], yticks=[])
    plt.imshow(im_convert(images[idx]))
    ax.set_title("{} ({})".format(str(classes[preds[idx].item()]), str(classes[labels[idx].item()])), color=("green" if preds[idx]==labels[idx] else "red"))
